# Autogenerated by Astropy's setup.py on 2019-10-08 00:29:49 UTC
from __future__ import unicode_literals
import datetime

version = "3.2.2"
githash = "bc6779fd4e6e174242a7fae599f7c42d2d60d0d5"


major = 3
minor = 2
bugfix = 2

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2019, 10, 8, 0, 29, 49)
debug = False

astropy_helpers_version = "3.2.1"
